""" Main class for DWM """

import sys
import collections

from .cleaning import DataLookup
from .cleaning import RegexLookup
from .cleaning import DeriveDataLookup
from .cleaning import DeriveDataCopyValue
from .cleaning import DeriveDataRegex
from .cleaning import IncludesLookup

##########################################################################
# Constants
##########################################################################

LOOKUP_TYPES = ['genericLookup', 'genericRegex', 'fieldSpecificRegex',
                'fieldSpecificLookup', 'normLookup', 'normRegex',
                'normIncludes']

DERIVE_TYPES = ['deriveValue', 'copyValue', 'deriveRegex', 'deriveIncludes']

DERIVE_OPTIONS = ['overwrite', 'blankIfNoMatch']

UDF_POSITIONS = ['beforeGenericValLookup', 'beforeGenericValRegex',
                 'beforeFieldSpecificLookup', 'beforeFieldSpecificRegex',
                 'beforeNormLookup', 'beforeNormRegex',
                 'beforeNormIncludes', 'beforeDerive', 'afterAll']

##########################################################################
# DWM
##########################################################################


class Dwm(object):
    """ class for DWM config """

    def __init__(self, name, mongo, fields=None, udfs=None):
        """
        Set configuration for DWM runtime

        :param str name: Name of configuration; for logging
        :param MongoClient mongo: MongoDB connection
        :param dict fields: dict of field configurations (dict)
        :param dict udfs: dict of udfs to run
        """

        if fields is None:
            fields = {}

        if udfs is None:
            udfs = {}

        # validate input values
        for field in fields:
            # lookup values
            for lookup in fields[field]['lookup']:
                if lookup not in LOOKUP_TYPES:
                    raise ValueError('Invalid lookup type %s' % lookup)

            # derive types and options
            for derive in fields[field]['derive']:
                if derive['type'] not in DERIVE_TYPES:
                    raise ValueError('Invalid derive type %s' % derive['type'])

                for opt in derive['options']:
                    if opt not in DERIVE_OPTIONS:
                        raise ValueError('Invalid derive option %s' % opt)

        for udf in udfs:
            if udf not in UDF_POSITIONS:
                raise ValueError('Invalid UDF position %s' % udf)

        self.name = name
        self.mongo = mongo
        self.fields = fields
        self.udfs = udfs

    def get_field_list(self):
        """
        Retrieve list of all fields currently configured
        """

        list_out = []
        for field in self.fields:
            list_out.append(field)

        return list_out

    @staticmethod
    def data_lookup_method(fields_list, mongo_db_obj, hist, record,
                           lookup_type):
        """
        Method to lookup the replacement value given a single input value from
        the same field.

        :param dict fields_list: Fields configurations
        :param MongoClient mongo_db_obj: MongoDB collection object
        :param dict hist: existing input of history values object
        :param dict record: values to validate
        :param str lookup_type: Type of lookup
        """

        if hist is None:
            hist = {}

        for field in record:

            if record[field] != '' and record[field] is not None:

                if field in fields_list:

                    if lookup_type in fields_list[field]['lookup']:

                        field_val_new, hist = DataLookup(
                            fieldVal=record[field],
                            db=mongo_db_obj,
                            lookupType=lookup_type,
                            fieldName=field,
                            histObj=hist)

                        record[field] = field_val_new

        return record, hist

    @staticmethod
    def data_regex_method(fields_list, mongo_db_obj, hist, record, lookup_type):
        """
        Method to lookup the replacement value based on regular expressions.

        :param dict fields_list: Fields configurations
        :param MongoClient mongo_db_obj: MongoDB collection object
        :param dict hist: existing input of history values object
        :param dict record: values to validate
        :param str lookup_type: Type of lookup
        """

        if hist is None:
            hist = {}

        for field in record:

            if record[field] != '' and record[field] is not None:

                if field in fields_list:

                    if lookup_type in fields_list[field]['lookup']:

                        field_val_new, hist = RegexLookup(
                            fieldVal=record[field],
                            db=mongo_db_obj,
                            fieldName=field,
                            lookupType=lookup_type,
                            histObj=hist)

                        record[field] = field_val_new

        return record, hist

    def _val_g_lookup(self, record, hist=None):
        """
        Perform generic validation lookup

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_lookup_method(fields_list=self.fields,
                                               mongo_db_obj=self.mongo,
                                               hist=hist,
                                               record=record,
                                               lookup_type='genericLookup')
        return record, hist

    def _val_g_regex(self, record, hist=None):
        """
        Perform generic validation regex

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_regex_method(fields_list=self.fields,
                                              mongo_db_obj=self.mongo,
                                              hist=hist,
                                              record=record,
                                              lookup_type='genericRegex')
        return record, hist

    def _val_fs_lookup(self, record, hist=None):
        """
        Perform field-specific validation lookup

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_lookup_method(fields_list=self.fields,
                                               mongo_db_obj=self.mongo,
                                               hist=hist,
                                               record=record,
                                               lookup_type=
                                               'fieldSpecificLookup')

        return record, hist

    def _val_fs_regex(self, record, hist=None):
        """
        Perform field-specific validation regex

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_regex_method(fields_list=self.fields,
                                              mongo_db_obj=self.mongo,
                                              hist=hist,
                                              record=record,
                                              lookup_type='fieldSpecificRegex')
        return record, hist

    def _norm_lookup(self, record, hist=None):
        """
        Perform generic validation lookup

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_lookup_method(fields_list=self.fields,
                                               mongo_db_obj=self.mongo,
                                               hist=hist,
                                               record=record,
                                               lookup_type='normLookup')
        return record, hist

    def _norm_regex(self, record, hist=None):
        """
        Perform generic validation regex

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        record, hist = self.data_regex_method(fields_list=self.fields,
                                              mongo_db_obj=self.mongo,
                                              hist=hist,
                                              record=record,
                                              lookup_type='normRegex')
        return record, hist

    def _norm_include(self, record, hist=None):
        """
        Normalization 'normIncludes' replace 'almost' values based on at least
        one of the following: includes strings, excludes strings, starts with
        string, ends with string

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        if hist is None:
            hist = {}

        for field in record:

            if record[field] != '' and record[field] is not None:

                if field in self.fields:

                    if 'normIncludes' in self.fields[field]['lookup']:

                        field_val_new, hist, _ = IncludesLookup(
                            fieldVal=record[field],
                            lookupType='normIncludes',
                            db=self.mongo,
                            fieldName=field,
                            histObj=hist)

                        record[field] = field_val_new

        return record, hist

    def _derive(self, record, hist=None):
        """
        Derivation filters like 'deriveValue' to replace given input values
        from one or more fields. In case 'copyValue' copy value to the target
        field from given an input value from one field. 'deriveRegex' replace
        given an input value from one field, derive target field value using
        regular expressions. If 'deriveIncludes' applies then given an input
        value from one field, derive target field based on at least one of the
        following: includes strings, excludes strings, starts with string,
        ends with string

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        def check_derive_options(option, derive_set_config):
            """
            Check derive option is exist into options list and return relevant
            flag.
            :param str option: drive options value
            :param list derive_set_config: options list
            :return boolean: True or False based on option exist into options
            list
            """

            return option in derive_set_config

        hist_obj = {}

        if hist is None:
            hist = {}

        for field in record:

            field_val_new = field_val = record[field]

            if field in self.fields:

                for derive_set in self.fields[field]['derive']:

                    check_match = False

                    derive_set_config = derive_set

                    if set.issubset(set(derive_set_config['fieldSet']),
                                    record.keys()):

                        # sorting here to ensure sub document match from
                        # query

                        derive_input = {val: record[val] for val in
                                        derive_set_config['fieldSet']}

                        if derive_set_config['type'] == 'deriveValue':

                            overwrite_flag = check_derive_options(
                                'overwrite',
                                derive_set_config["options"])

                            blank_if_no_match_flag = check_derive_options(
                                'blankIfNoMatch',
                                derive_set_config["options"])

                            field_val_new, hist_obj, check_match = \
                                DeriveDataLookup(
                                    fieldName=field,
                                    db=self.mongo,
                                    deriveInput=derive_input,
                                    overwrite=overwrite_flag,
                                    fieldVal=record[field],
                                    histObj=hist,
                                    blankIfNoMatch=blank_if_no_match_flag)

                        elif derive_set_config['type'] == 'copyValue':

                            overwrite_flag = check_derive_options(
                                'overwrite',
                                derive_set_config["options"])

                            field_val_new, hist_obj, check_match = \
                                DeriveDataCopyValue(
                                    fieldName=field,
                                    deriveInput=derive_input,
                                    overwrite=overwrite_flag,
                                    fieldVal=record[field],
                                    histObj=hist)

                        elif derive_set_config['type'] == 'deriveRegex':

                            overwrite_flag = check_derive_options(
                                'overwrite',
                                derive_set_config["options"])

                            blank_if_no_match_flag = check_derive_options(
                                'blankIfNoMatch',
                                derive_set_config["options"])

                            field_val_new, hist_obj, check_match = \
                                DeriveDataRegex(
                                    fieldName=field,
                                    db=self.mongo,
                                    deriveInput=derive_input,
                                    overwrite=overwrite_flag,
                                    fieldVal=record[field],
                                    histObj=hist,
                                    blankIfNoMatch=blank_if_no_match_flag)

                        elif derive_set_config['type'] == 'deriveIncludes':

                            overwrite_flag = check_derive_options(
                                'overwrite',
                                derive_set_config["options"])

                            blank_if_no_match_flag = check_derive_options(
                                'blankIfNoMatch',
                                derive_set_config["options"])

                            field_val_new, hist_obj, check_match = \
                                IncludesLookup(
                                    fieldVal=record[field],
                                    lookupType='deriveIncludes',
                                    deriveFieldName= \
                                        derive_set_config['fieldSet'][0],
                                    deriveInput=derive_input,
                                    db=self.mongo,
                                    fieldName=field,
                                    histObj=hist,
                                    overwrite=overwrite_flag,
                                    blankIfNoMatch=blank_if_no_match_flag)

                    if check_match or field_val_new != field_val:
                        record[field] = field_val_new
                        break

        return record, hist_obj

    def _apply_udfs(self, record, hist, udf_type):
        """
        Excute user define processes, user-defined functionalty is designed to
        applyies custome trasformations to data.

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        """

        def function_executor(func, *args):
            """
            Execute user define function
            :param python method func: Function obj
            :param methods arguments args: Function arguments
            """

            result, result_hist = func(*args)

            return result, result_hist

        if udf_type in self.udfs:

            cust_function_od_obj = collections.OrderedDict(
                sorted(
                    self.udfs[udf_type].items()
                )
            )

            for cust_function in cust_function_od_obj:

                record, hist = function_executor(
                    cust_function_od_obj[cust_function],
                    record,
                    hist
                )

        return record, hist

    def run(self, record, hist=None):
        """
        By passing the input record to be cleaned, and returns the input record
        after cleaning with history(dictionary).

        :param dict record: dictionary of values to validate
        :param dict hist: existing input of history values
        :return:
        """

        if hist is None:
            hist = {}

        if record:
            # Run user-defined functions for beforeGenericValLookup
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeGenericValLookup')

            # Run generic validation lookup
            record, hist = self._val_g_lookup(record=record, hist=hist)

            # Run user-defined functions for beforeGenericValRegex
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeGenericValRegex')

            # Run generic validation regex
            record, hist = self._val_g_regex(record=record, hist=hist)

            # Run user-defined functions for beforeFieldSpecificLookup
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeFieldSpecificLookup')

            # Run field-specific validation lookup
            record, hist = self._val_fs_lookup(record=record, hist=hist)

            # Run user-defined functions for beforeFieldSpecificLookup
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeFieldSpecificRegex')

            # Run field-specific validation regex
            record, hist = self._val_fs_regex(record=record, hist=hist)

            # Run user-defined functions for beforeNormLookup
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeNormLookup')

            # Run normalization lookup
            record, hist = self._norm_lookup(record=record, hist=hist)

            # Run user-defined functions for beforeNormRegex
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeNormRegex')

            # Run normalization regex
            record, hist = self._norm_regex(record=record, hist=hist)

            # Run user-defined functions for beforeNormIncludes
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeNormIncludes')

            # Run normalization includes
            record, hist = self._norm_include(record=record, hist=hist)

            # Run user-defined functions for beforeNormIncludes
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='beforeDerive')

            # Fill gaps / refresh derived data
            record, hist = self._derive(record=record, hist=hist)

            # Run user-defined functions for beforeNormIncludes
            record, hist = self._apply_udfs(record=record,
                                            hist=hist,
                                            udf_type='afterAll')

            return record, hist

        return None, hist



record_lookupAll_genericLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_lookupAll_genericLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_lookupAll_genericLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_lookupAll_fieldSpecificLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_lookupAll_fieldSpecificLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_lookupAll_fieldSpecificLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_lookupAll_normLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_lookupAll_normLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_lookupAll_normLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_regexAll_genericRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_regexAll_genericRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_regexAll_genericRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_regexAll_fieldSpecificRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_regexAll_fieldSpecificRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_regexAll_fieldSpecificRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_regexAll_normRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

record_regexAll_normRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

record_regexAll_normRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

record_normIncludes_included_caught = [{"emailAddress": "test@test.com", "field1": "findgoodinvaluejunk"}]

record_normIncludes_included_uncaught = [{"emailAddress": "test@test.com", "field1": "nothere"}]

record_normIncludes_excluded_caught = [{"emailAddress": "test@test.com", "field1": "findgoodinvaluejunk butstuffnobad"}]

record_normIncludes_excluded_uncaught = [{"emailAddress": "test@test.com", "field1": "findgoodinvaluejunk uncaught"}]

record_normIncludes_begins_caught = [{"emailAddress": "test@test.com", "field1": "abcdefg"}]

record_normIncludes_begins_uncaught = [{"emailAddress": "test@test.com", "field1": "hijklmnop"}]

record_normIncludes_ends_caught = [{"emailAddress": "test@test.com", "field1": "qrstuvwxyz"}]

record_normIncludes_ends_uncaught = [{"emailAddress": "test@test.com", "field1": "notalpha"}]

record_normIncludes_notChecked = [{"emailAddress": "test@test.com", "field2": "doesnotmatter"}]

record_deriveAll_deriveValue_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findthis"}]

record_derive_sort = [{"emailAddress": "test@test.com", "field1": "", "field3": "findthis", "field4": "nofindthis"}]

record_deriveAll_deriveValue_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findthis"}]

record_deriveAll_deriveValue_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "youwillnotfindthis"}]

record_deriveAll_deriveValue_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "dontfindthis"}]

record_deriveAll_deriveValue_notChecked = [{"emailAddress": "test@test.com", "field3": "", "field2": "findthis"}]

record_deriveAll_copyValue = [{"emailAddress": "test@test.com", "field1": "", "field2": "newvalue"}]

record_deriveAll_deriveRegex_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findthis"}]

record_deriveAll_deriveRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "dontfindthis"}]

record_deriveAll_deriveRegex_notChecked = [{"emailAddress": "test@test.com", "field3": "", "field2": "findthis"}]

record_deriveAll_deriveRegex_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findthis"}]

record_deriveAll_deriveRegex_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "youwillnotfindthis"}]

record_deriveAll_deriveIncludes_included_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk"}]

record_deriveAll_deriveIncludes_included_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "nothere"}]

record_deriveAll_deriveIncludes_excluded_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk butstuffnobad"}]

record_deriveAll_deriveIncludes_excluded_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk uncaught"}]

record_deriveAll_deriveIncludes_begins_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "abcdefg"}]

record_deriveAll_deriveIncludes_begins_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "hijklmnop"}]

record_deriveAll_deriveIncludes_ends_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "qrstuvwxyz"}]

record_deriveAll_deriveIncludes_ends_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "notalpha"}]

record_deriveAll_deriveIncludes_notChecked = [{"emailAddress": "test@test.com", "field2": "", "field3": "doesnotmatter"}]

record_deriveAll_deriveIncludes_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findgoodinvaluejunk"}]

record_deriveAll_deriveIncludes_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "nothere"}]
#

history_genericLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_genericLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_genericLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_fieldSpecificLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_fieldSpecificLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_fieldSpecificLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_normLookup_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_normLookup_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_normLookup_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_genericRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_genericRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_genericRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_fieldSpecificRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_fieldSpecificRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_fieldSpecificRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_normRegex_caught = [{"emailAddress": "test@test.com", "field1": "badvalue"}]

history_normRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "badvalue-uncaught"}]

history_normRegex_notChecked = [{"emailAddress": "test@test.com", "field2": "badvalue"}]

history_normIncludes_included_caught = [{"emailAddress": "test@test.com", "field1": "findgoodinvaluejunk"}]

history_normIncludes_included_uncaught = [{"emailAddress": "test@test.com", "field1": "nothere"}]

history_normIncludes_notChecked = [{"emailAddress": "test@test.com", "field2": "doesnotmatter"}]

history_deriveValue_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findthis"}]

history_deriveValue_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findthis"}]

history_deriveValue_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "youwillnotfindthis"}]

history_deriveValue_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "dontfindthis"}]

history_deriveValue_notChecked = [{"emailAddress": "test@test.com", "field3": "", "field2": "findthis"}]

history_copyValue = [{"emailAddress": "test@test.com", "field1": "", "field2": "newvalue"}]

history_deriveRegex_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findthis"}]

history_deriveRegex_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "dontfindthis"}]

history_deriveRegex_notChecked = [{"emailAddress": "test@test.com", "field3": "", "field2": "findthis"}]

history_deriveRegex_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findthis"}]

history_deriveRegex_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "youwillnotfindthis"}]

history_deriveIncludes_included_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk"}]

history_deriveIncludes_included_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "nothere"}]

history_deriveIncludes_excluded_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk butstuffnobad"}]

history_deriveIncludes_excluded_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk uncaught"}]

history_deriveIncludes_begins_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "abcdefg"}]

history_deriveIncludes_begins_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "hijklmnop"}]

history_deriveIncludes_ends_caught = [{"emailAddress": "test@test.com", "field1": "", "field2": "qrstuvwxyz"}]

history_deriveIncludes_ends_uncaught = [{"emailAddress": "test@test.com", "field1": "", "field2": "notalpha"}]

history_deriveIncludes_notChecked = [{"emailAddress": "test@test.com", "field2": "", "field3": "doesnotmatter"}]

history_deriveIncludes_overwriteFalse = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "findgoodinvaluejunk"}]

history_deriveIncludes_blankIfNoMatch = [{"emailAddress": "test@test.com", "field1": "oldvalue", "field2": "nothere"}]

history_deriveIncludes_deriveCheckMatch = [{"emailAddress": "test@test.com", "field1": "", "field2": "findgoodinvaluejunk", "field3":"notright"}]

record_returnHistoryId_False = [{"emailAddress": "test@test.com"}]

record_writeContactHistory_False = [{"emailAddress": "test@test.com"}]

record_writeContactHistory_writeConfig = [{"emailAddress": "test@test.com"}]

record_writeContactHistory_historyCurrent1 = [{"emailAddress": "test@test.com"}]

record_writeContactHistory_historyCurrent2 = [{"emailAddress": "test@test.com"}]

record_configDoesNotExist = [{"emailAddress": "test@test.com"}]

record_udf_beforeGenericValidation = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeGenericRegex = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeFieldSpecificValidation = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeFieldSpecificRegex = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeNormalization = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeNormalizationRegex = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeNormalizationIncludes = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_beforeDeriveData = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_afterProcessing = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_sort = [{"emailAddress": "test@test.com", "field1": ""}]

record_udf_afterProcessing_invalidFcn = [{"emailAddress": "test@test.com", "field1": ""}]

record_dwmAll_noConfig = [{"emailAddress": "test@test.com"}]



""" test derive includes case of derive function"""

import mongomock

from dwm import Dwm
from .test_deriveIncludes import deriveIncludes


# Setup mongomock db
DB = mongomock.MongoClient().db

DB.deriveIncludes.insert_many(deriveIncludes)

# Setup Dwm instance
FIELDS = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "deriveIncludes",
                "fieldSet": ["field2"],
                "options": ["overwrite"]
            }
        ]
    }
}


DWM = Dwm(name='test', mongo=DB, fields=FIELDS)

FIELDS_OVERWRITE_FALSE = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "deriveIncludes",
                "fieldSet": ["field2"],
                "options": []
            }
        ]
    }
}

DWM_OVERWRITE_FALSE_OBJ = Dwm(name='test', mongo=DB,
                              fields=FIELDS_OVERWRITE_FALSE)

FIELDS_BLANK_IF_NO_MATCH = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "deriveIncludes",
                "fieldSet": ["field2"],
                "options": ["overwrite", "blankIfNoMatch"]
            }
        ]
    }
}

DWM_BLANK_IF_NO_MATCH_OBJ = Dwm(name='test', mongo=DB,
                                fields=FIELDS_BLANK_IF_NO_MATCH)


# Let the testing begin
def test_derive_includes_included_caught():
    """ Ensure derive includes caught """
    rec = {"emailAddress": "test@test.com", "field1": "",
           "field2": "findgoodinvaluejunk"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_derive_includes_included_uncaught():
    """ Ensure derive includes uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "", "field2": "nothere"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_derive_includes_excluded_caught():
    """ Ensure derive includes excluded caught """
    rec = {"emailAddress": "test@test.com", "field1": "",
           "field2": "findgoodinvaluejunk butstuffnobad"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_derive_includes_excluded_uncaught():
    """ Ensure derive includes excluded uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "",
           "field2": "findgoodinvaluejunk uncaught"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_derive_includes_begins_caught():
    """ Ensure derive includes begins caught """
    rec = {"emailAddress": "test@test.com", "field1": "", "field2": "abcdefg"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_derive_includes_begins_uncaught():
    """ Ensure derive includes begins uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "", "field2": "hijklmnop"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_derive_includes_ends_caught():
    """ Ensure derive includes ends caught """
    rec = {"emailAddress": "test@test.com", "field1": "",
           "field2": "qrstuvwxyz"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_derive_includes_ends_uncaught():
    """ Ensure derive includes ends uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "", "field2": "notalpha"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_derive_includes_not_checked():
    """ Ensure derive includes not checked """
    rec = {"emailAddress": "test@test.com", "field2": "",
           "field3": "doesnotmatter"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field2'] != 'goodvalue'


def test_derive_includes_over_write_false():
    """ Ensure derive includes over write false"""
    rec = {"emailAddress": "test@test.com", "field1": "oldvalue",
           "field2": "findgoodinvaluejunk"}
    rec_out, _ = DWM_OVERWRITE_FALSE_OBJ._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'oldvalue'


def test_derive_includes_blank_if_no_match():
    """ Ensure derive includes blank if no match"""
    rec = {"emailAddress": "test@test.com", "field1": "oldvalue",
           "field2": "nothere"}
    rec_out, _ = DWM_BLANK_IF_NO_MATCH_OBJ._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == ''



configs = [

    {
        "configName": "test_lookupAll_genericLookup",
        "fields": {
            "field1": {
                "lookup": ["genericLookup"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_lookupAll_fieldSpecificLookup",
        "fields": {
            "field1": {
                "lookup": ["fieldSpecificLookup"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_lookupAll_normLookup",
        "fields": {
            "field1": {
                "lookup": ["normLookup"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_normIncludes",
        "fields": {
            "field1": {
                "lookup": ["normIncludes"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_regexAll_genericRegex",
        "fields": {
            "field1": {
                "lookup": ["genericRegex"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_regexAll_fieldSpecificRegex",
        "fields": {
            "field1": {
                "lookup": ["fieldSpecificRegex"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_regexAll_normRegex",
        "fields": {
            "field1": {
                "lookup": ["normRegex"],
                "derive": {}
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveValue",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveValue",
                        "fieldSet": ["field2"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveValue_blankIfNoMatch",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveValue",
                        "fieldSet": ["field2"],
                        "options": ["overwrite", "blankIfNoMatch"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_derive_sort",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "2": {
                        "type": "deriveValue",
                        "fieldSet": ["field4"],
                        "options": ["overwrite"]
                    },
                    "1": {
                        "type": "deriveValue",
                        "fieldSet": ["field3"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_copyValue",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "copyValue",
                        "fieldSet": ["field2"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveRegex",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveRegex",
                        "fieldSet": ["field2"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveIncludes",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveIncludes",
                        "fieldSet": ["field2"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveIncludes_deriveCheckMatch",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveIncludes",
                        "fieldSet": ["field2"],
                        "options": ["overwrite"]
                    },
                    "2": {
                        "type": "copyValue",
                        "fieldSet": ["field3"],
                        "options": ["overwrite"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
        },
        "history": {
            "writeContactHistory": True,
            "returnHistoryId": True,
            "returnHistoryField": "historyId",
            "histIdField": {"name": "field1", "value": "field1"}
        }
    },
    {
        "configName": "test_deriveAll_deriveIncludes_overwriteFalse",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveIncludes",
                        "fieldSet": ["field2"],
                        "options": []
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveIncludes_blankIfNoMatch",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveIncludes",
                        "fieldSet": ["field2"],
                        "options": ["overwrite", "blankIfNoMatch"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveRegex_blankIfNoMatch",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveRegex",
                        "fieldSet": ["field2"],
                        "options": ["overwrite", "blankIfNoMatch"]
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveValue_overwriteFalse",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveValue",
                        "fieldSet": ["field2"],
                        "options": []
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_deriveAll_deriveRegex_overwriteFalse",
        "fields": {
            "field1": {
                "lookup": [],
                "derive": {
                    "1": {
                        "type": "deriveRegex",
                        "fieldSet": ["field2"],
                        "options": []
                    }
                }
            }
        },
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_returnHistoryId_False",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": False,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_writeContactHistory_False",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": False,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_writeContactHistory_writeConfig",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeGenericValidation",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {
                "1": "ex_udf"
            },
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeGenericRegex",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {
                "1": "ex_udf"
            },
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeFieldSpecificValidation",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {
                "1": "ex_udf"
            },
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeFieldSpecificRegex",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {
                "1": "ex_udf"
            },
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeNormalization",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {
                "1": "ex_udf"
            },
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeNormalizationRegex",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {
                "1": "ex_udf"
            },
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeNormalizationIncludes",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {
                "1": "ex_udf"
            },
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_beforeDeriveData",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {
                "1": "ex_udf"
            },
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_afterProcessing",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {
                "1": "ex_udf"
            }
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_sort",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {
                "3": "ex_udf",
                "1": "sort_udf_1",
                "2": "sort_udf_2"
            },
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {}
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    },
    {
        "configName": "test_udf_afterProcessing_invalidFcn",
        "fields": {},
        "userDefinedFunctions": {
            "beforeGenericValidation": {},
            "beforeGenericRegex": {},
            "beforeFieldSpecificValidation": {},
            "beforeFieldSpecificRegex": {},
            "beforeNormalization": {},
            "beforeNormalizationRegex": {},
            "beforeNormalizationIncludes": {},
            "beforeDeriveData": {},
            "afterProcessing": {
                "1": "bad_udf"
            }
      },
        "history": {
          "writeContactHistory": True,
          "returnHistoryId": True,
          "returnHistoryField": "historyId",
          "histIdField": {"name": "emailAddress", "value": "emailAddress"}
        }
    }
]



""" test normalization validation lookup function """

import mongomock
#from mock import patch
#from nose.tools import raises

from dwm import Dwm

# Setup mongomock db

DB = mongomock.MongoClient().db

DB.normLookup.insert({"fieldName": "field1", "find": "BADVALUE", "replace": "goodvalue"})
DB.normLookup.insert({"fieldName": "field2", "find": "BADVALUE", "replace": "goodvalue"})

# Setup Dwm instance

FIELDS = {
    'field1': {
        'lookup': ['normLookup'],
        'derive': []
    },
    'field2': {
        'lookup': ['normLookup'],
        'derive': []
    }
}

DWM = Dwm(name='test', mongo=DB, fields=FIELDS)

# Let the testing begin


def test_dwm_vnorm_lup_bad():
    """ Ensure field-specific lookup occurs """
    rec = {'field1': 'BADVALUE'}
    rec_out, _ = DWM._norm_lookup(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': 'goodvalue'}


def test_dwm_vnorm_lup_good():
    """ Ensure good value not cleared by fs """
    rec = {'field1': 'GOODVALUE'}
    rec_out, _ = DWM._norm_lookup(rec, {}) #pylint: disable=W0212
    assert rec_out == rec


def test_dwm_vnorm_lup_badcln():
    """ Ensure basic fs lookup occurs and cleans value before """
    rec = {'field1': '  badvalue\r\n  '}
    rec_out, _ = DWM._norm_lookup(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': 'goodvalue'}


def test_dwm_vnorm_lup_badmulti():
    """ Ensure fs lookup occurs on every field in config """
    rec = {'field1': 'BADVALUE', 'field2': 'BADVALUE'}
    rec_out, _ = DWM._norm_lookup(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': 'goodvalue', 'field2': 'goodvalue'}


def test_dwm_vnorm_lup_leave():
    """ Ensure fs lookup does not occur on field not in config """
    rec = {'field1': 'BADVALUE', 'field3': 'BADVALUE'}
    rec_out, _ = DWM._norm_lookup(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': 'goodvalue', 'field3': 'BADVALUE'}



""" test generic validation regex function """

import mongomock
#from mock import patch
#from nose.tools import raises

from dwm import Dwm

# Setup mongomock db

DB = mongomock.MongoClient().db

DB.genericRegex.insert({"pattern": r"^badvalue$"})

# Setup Dwm instance

FIELDS = {
    'field1': {
        'lookup': ['genericRegex'],
        'derive': []
    },
    'field2': {
        'lookup': ['genericRegex'],
        'derive': []
    }
}

DWM = Dwm(name='test', mongo=DB, fields=FIELDS)

# Let the testing begin

def test_dwm_vg_reg_bad():
    """ Ensure generic regex occurs """
    rec = {'field1': 'BADVALUE'}
    rec_out, _ = DWM._val_g_regex(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': ''}

def test_dwm_vg_reg_good():
    """ Ensure good value not cleared """
    rec = {'field1': 'GOODVALUE'}
    rec_out, _ = DWM._val_g_regex(rec, {}) #pylint: disable=W0212
    assert rec_out == rec

def test_dwm_vg_reg_badcln():
    """ Ensure basic regex occurs and cleans value before """
    rec = {'field1': '  badvalue\r\n  '}
    rec_out, _ = DWM._val_g_regex(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': ''}

def test_dwm_vg_reg_badmulti():
    """ Ensure regex occurs on every field in config """
    rec = {'field1': 'BADVALUE', 'field2': 'BADVALUE'}
    rec_out, _ = DWM._val_g_regex(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': '', 'field2': ''}

def test_dwm_vg_reg_leave():
    """ Ensure regex does not occur on field not in config """
    rec = {'field1': 'BADVALUE', 'field3': 'BADVALUE'}
    rec_out, _ = DWM._val_g_regex(rec, {}) #pylint: disable=W0212
    assert rec_out == {'field1': '', 'field3': 'BADVALUE'}



""" test norm include function """

import mongomock

from dwm import Dwm
from .test_normIncludes import normIncludes


# Setup mongomock db
DB = mongomock.MongoClient().db

DB.normIncludes.insert_many(normIncludes)

# Setup Dwm instance
FIELDS = {
    "field1": {
        "lookup": ["normIncludes"],
        "derive": []
    }
}

DWM = Dwm(name='test', mongo=DB, fields=FIELDS)

FIELDS_OVERWRITE_FALSE = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "deriveValue",
                "fieldSet": ["field2"],
                "options": []
            }
        ]
    }
}

DWM_OVERWRITE_FALSE_OBJ = Dwm(name='test', mongo=DB,
                              fields=FIELDS_OVERWRITE_FALSE)

FIELDS_BLANK_IF_NO_MATCH = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "deriveValue",
                "fieldSet": ["field2"],
                "options": ["overwrite", "blankIfNoMatch"]
            }
        ]
    }
}

DWM_BLANK_IF_NO_MATCH_OBJ = Dwm(name='test', mongo=DB,
                                fields=FIELDS_BLANK_IF_NO_MATCH)


# Let the testing begin
def test_norm_includes_included_caught():
    """ Ensure norm includes included caught """
    rec = {"emailAddress": "test@test.com", "field1": "findgoodinvaluejunk"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_norm_includes_included_uncaught():
    """ Ensure norm includes included uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "nothere"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_norm_includes_excluded_caught():
    """ Ensure norm includes excluded caught """
    rec = {"emailAddress": "test@test.com",
           "field1": "findgoodinvaluejunk butstuffnobad"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_norm_includes_excluded_uncaught():
    """ Ensure norm includes excluded uncaught """
    rec = {"emailAddress": "test@test.com",
           "field1": "findgoodinvaluejunk uncaught"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_norm_includes_begins_caught():
    """ Ensure norm includes begins caught """
    rec = {"emailAddress": "test@test.com", "field1": "abcdefg"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_norm_includes_begins_uncaught():
    """ Ensure norm includes begins uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "hijklmnop"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_norm_includes_ends_caught():
    """ Ensure norm includes ends caught """
    rec = {"emailAddress": "test@test.com", "field1": "qrstuvwxyz"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_norm_includes_ends_uncaught():
    """ Ensure norm includes ends uncaught """
    rec = {"emailAddress": "test@test.com", "field1": "notalpha"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] != 'goodvalue'


def test_norm_includes_not_checked():
    """ Ensure norm includes not checked """
    rec = {"emailAddress": "test@test.com", "field2": "doesnotmatter"}
    rec_out, _ = DWM._norm_include(rec, {})  # pylint: disable=W0212
    assert rec_out['field2'] != 'goodvalue'



"""
These set of functions are responsible for determining what the new value of a
field should be, in most cases based on a lookup against MongoDB.
"""

import re
import warnings

from collections import OrderedDict

from .helpers import _CollectHistory_
from .helpers import _CollectHistoryAgg_
from .helpers import _DataClean_


def DataLookup(fieldVal, db, lookupType, fieldName, histObj={}):
    """
    Return new field value based on single-value lookup against MongoDB

    :param string fieldVal: input value to lookup
    :param MongoClient db: MongoClient instance connected to MongoDB
    :param string lookupType: Type of lookup to perform/MongoDB collection name.
           One of 'genericLookup', 'fieldSpecificLookup', 'normLookup'
    :param string fieldName: Field name to query against
    :param dict histObj: History object to which changes should be appended
    """

    if lookupType == 'genericLookup':
        lookup_dict = {"find": _DataClean_(fieldVal)}
    elif lookupType in ['fieldSpecificLookup', 'normLookup']:
        lookup_dict = {"fieldName": fieldName, "find": _DataClean_(fieldVal)}
    else:
        raise ValueError("Invalid lookupType")

    field_val_new = fieldVal

    coll = db[lookupType]

    l_val = coll.find_one(lookup_dict, ['replace'])

    if l_val:
        field_val_new = l_val['replace'] if 'replace' in l_val else ''

    change = _CollectHistory_(lookupType=lookupType, fromVal=fieldVal,
                              toVal=field_val_new)

    hist_obj_upd = _CollectHistoryAgg_(contactHist=histObj,
                                       fieldHistObj=change,
                                       fieldName=fieldName)

    return field_val_new, hist_obj_upd


def IncludesLookup(fieldVal, lookupType, db, fieldName, deriveFieldName='',
                   deriveInput={}, histObj={}, overwrite=False,
                   blankIfNoMatch=False):
    """
    Return new field value based on whether or not original value includes AND
    excludes all words in a comma-delimited list queried from MongoDB

    :param string fieldVal: input value to lookup
    :param string lookupType: Type of lookup to perform/MongoDB collection name.
           One of 'normIncludes', 'deriveIncludes'
    :param MongoClient db: MongoClient instance connected to MongoDB
    :param string fieldName: Field name to query against
    :param string deriveFieldName: Field name from which to derive value
    :param dict deriveInput: Values to perform lookup against:
           {"deriveFieldName": "deriveVal1"}
    :param dict histObj: History object to which changes should be appended
    :param bool overwrite: Should an existing field value be replaced
    :param bool blankIfNoMatch: Should field value be set to blank if
           no match is found
    """

    lookup_dict = {
        'fieldName': fieldName
    }

    if lookupType == 'normIncludes':
        field_val_clean = _DataClean_(fieldVal)

    elif lookupType == 'deriveIncludes':

        if deriveFieldName == '' or deriveInput == {}:
            raise ValueError("for 'deriveIncludes' must specify both \
                              'deriveFieldName' and 'deriveInput'")

        lookup_dict['deriveFieldName'] = deriveFieldName
        field_val_clean = _DataClean_(deriveInput[list(deriveInput.keys())[0]])
    else:
        raise ValueError("Invalid lookupType")

    field_val_new = fieldVal
    check_match = False
    using = {}

    coll = db[lookupType]

    inc_val = coll.find(lookup_dict, ['includes', 'excludes', 'begins', 'ends',
                                      'replace'])

    if inc_val and (lookupType == 'normIncludes' or
                    (lookupType == 'deriveIncludes' and
                     (overwrite or fieldVal == ''))):

        for row in inc_val:

            try:

                if (row['includes'] != '' or
                        row['excludes'] != '' or
                        row['begins'] != '' or
                        row['ends'] != ''):

                    if all((a in field_val_clean)
                           for a in row['includes'].split(",")):

                        if all((b not in field_val_clean)
                               for b in row['excludes'].split(",")) \
                                or row['excludes'] == '':

                            if field_val_clean.startswith(row['begins']):

                                if field_val_clean.endswith(row['ends']):

                                    field_val_new = row['replace']

                                    if lookupType == 'deriveIncludes':
                                        using[deriveFieldName] = deriveInput

                                    using['includes'] = row['includes']
                                    using['excludes'] = row['excludes']
                                    using['begins'] = row['begins']
                                    using['ends'] = row['ends']

                                    check_match = True

                                    break

            except KeyError as Key_error_obj:
                warnings.warn('schema error', Key_error_obj)

        if inc_val:
            inc_val.close()

    if (field_val_new == fieldVal and blankIfNoMatch and
            lookupType == 'deriveIncludes'):
        field_val_new = ''
        using['blankIfNoMatch'] = 'no match found'

    change = _CollectHistory_(lookupType=lookupType, fromVal=fieldVal,
                              toVal=field_val_new, using=using)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change,
                                     fieldName=fieldName)

    return field_val_new, histObjUpd, check_match


def RegexLookup(fieldVal, db, fieldName, lookupType, histObj={}):
    """
    Return a new field value based on match against regex queried from MongoDB

    :param string fieldVal: input value to lookup
    :param MongoClient db: MongoClient instance connected to MongoDB
    :param string lookupType: Type of lookup to perform/MongoDB collection name.
            One of 'genericRegex', 'fieldSpecificRegex', 'normRegex'
    :param string fieldName: Field name to query against
    :param dict histObj: History object to which changes should be appended
    """

    if lookupType == 'genericRegex':
        lookup_dict = {}
    elif lookupType in ['fieldSpecificRegex', 'normRegex']:
        lookup_dict = {"fieldName": fieldName}
    else:
        raise ValueError("Invalid type")

    field_val_new = fieldVal
    pattern = ''

    coll = db[lookupType]

    re_val = coll.find(lookup_dict, ['pattern', 'replace'])

    for row in re_val:

        try:
            match = re.match(row['pattern'], _DataClean_(field_val_new),
                             flags=re.IGNORECASE)

            if match:

                if 'replace' in row:
                    field_val_new = re.sub(row['pattern'], row['replace'],
                                           _DataClean_(field_val_new),
                                           flags=re.IGNORECASE)
                else:
                    field_val_new = re.sub(row['pattern'], '',
                                           _DataClean_(field_val_new),
                                           flags=re.IGNORECASE)

                pattern = row['pattern']
                break

        except KeyError as Key_error_obj:
            warnings.warn('schema error', Key_error_obj)

    if re_val:
        re_val.close()

    change = _CollectHistory_(lookupType=lookupType, fromVal=fieldVal,
                              toVal=field_val_new, pattern=pattern)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change,
                                     fieldName=fieldName)

    return field_val_new, histObjUpd


def DeriveDataLookup(fieldName, db, deriveInput, overwrite=True, fieldVal='',
                     histObj={}, blankIfNoMatch=False):
    """
    Return new field value based on single or multi-value lookup against MongoDB

    :param string fieldName: Field name to query against
    :param MongoClient db: MongoClient instance connected to MongoDB
    :param dict deriveInput: Values to perform lookup against:
           {"lookupField1": "lookupVal1", "lookupField2": "lookupVal2"}
    :param bool overwrite: Should an existing field value be replaced
    :param string fieldVal: Current field value
    :param dict histObj: History object to which changes should be appended
    :param bool blankIfNoMatch: Should field value be set to blank
           if no match is found
    """

    lookup_vals = OrderedDict()

    for val in sorted(deriveInput.keys()):
        lookup_vals[val] = _DataClean_(deriveInput[val])

    lookup_dict = {
        'fieldName': fieldName,
        'lookupVals': lookup_vals
    }

    coll = db['deriveValue']

    l_val = coll.find_one(lookup_dict, ['value'])

    field_val_new = fieldVal

    derive_using = deriveInput

    # If match found return True else False
    check_match = True if l_val else False

    if l_val and (overwrite or (fieldVal == '')):

        try:
            field_val_new = l_val['value']
        except KeyError as Key_error_obj:
            warnings.warn('schema error', Key_error_obj)

    elif blankIfNoMatch and not l_val:

        field_val_new = ''
        derive_using = {'blankIfNoMatch': 'no match found'}

    change = _CollectHistory_(lookupType='deriveValue', fromVal=fieldVal,
                              toVal=field_val_new, using=derive_using)

    hist_obj_upd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change,
                                       fieldName=fieldName)

    return field_val_new, hist_obj_upd, check_match


def DeriveDataCopyValue(fieldName, deriveInput, overwrite, fieldVal, histObj={}):
    """
    Return new value based on value from another field

    :param string fieldName: Field name to query against
    :param dict deriveInput: Values to perform lookup against:
           {"copyField1": "copyVal1"}
    :param bool overwrite: Should an existing field value be replaced
    :param string fieldVal: Current field value
    :param dict histObj: History object to which changes should be appended
    """

    if len(deriveInput) > 1:
        raise Exception("more than one field/value in deriveInput")

    field_val_new = fieldVal

    row = list(deriveInput.keys())[0]

    if deriveInput[row] != '' and (overwrite or (fieldVal == '')):
        field_val_new = deriveInput[row]
        check_match = True
    else:
        check_match = False

    change = _CollectHistory_(lookupType='copyValue', fromVal=fieldVal,
                              toVal=field_val_new, using=deriveInput)

    hist_obj_upd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change,
                                       fieldName=fieldName)

    return field_val_new, hist_obj_upd, check_match


def DeriveDataRegex(fieldName, db, deriveInput, overwrite, fieldVal, histObj={},
                    blankIfNoMatch=False):
    """
    Return a new field value based on match (of another field) against regex
    queried from MongoDB

    :param string fieldName: Field name to query against
    :param MongoClient db: MongoClient instance connected to MongoDB
    :param dict deriveInput: Values to perform lookup against:
           {"lookupField1": "lookupVal1"}
    :param bool overwrite: Should an existing field value be replaced
    :param string fieldVal: Current field value
    :param dict histObj: History object to which changes should be appended
    :param bool blankIfNoMatch: Should field value be set to blank
           if no match is found
    """

    if len(deriveInput) > 1:
        raise Exception("more than one value in deriveInput")

    field_val_new = fieldVal
    check_match = False

    # derive_using = deriveInput

    row = list(deriveInput.keys())[0]

    pattern = ''

    if deriveInput[row] != '' and (overwrite or (fieldVal == '')):

        lookup_dict = {
            'deriveFieldName': row,
            'fieldName': fieldName
        }

        coll = db['deriveRegex']

        re_val = coll.find(lookup_dict, ['pattern', 'replace'])

        for l_val in re_val:

            try:

                match = re.match(l_val['pattern'],
                                 _DataClean_(deriveInput[row]),
                                 flags=re.IGNORECASE)

                if match:

                    field_val_new = re.sub(l_val['pattern'], l_val['replace'],
                                           _DataClean_(deriveInput[row]),
                                           flags=re.IGNORECASE)

                    pattern = l_val['pattern']

                    check_match = True
                    break

            except KeyError as key_error_obj:
                warnings.warn('schema error', key_error_obj)

        if re_val:
            re_val.close()

        if field_val_new == fieldVal and blankIfNoMatch:
            field_val_new = ''
            pattern = 'no matching pattern'
            # derive_using = {"blankIfNoMatch": "no match found"}

    change = _CollectHistory_(lookupType='deriveRegex', fromVal=fieldVal,
                              toVal=field_val_new, using=deriveInput,
                              pattern=pattern)

    hist_obj_upd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change,
                                       fieldName=fieldName)

    return field_val_new, hist_obj_upd, check_match



deriveValue = [
    {"fieldName": "field1", "lookupVals": {"field2": "FINDTHIS"}, "value": "newvalue"},
    {"fieldName": "field1", "lookupVals": {"field4": "NOFINDTHIS"}, "value": "incorrectvalue"},
    {"fieldName": "field1", "lookupVals": {"field3": "FINDTHIS"}, "value": "correctvalue"},
    {"fieldName": "field100", "lookupVals": {"field101": "ABC"}}
]



""" test copy value case of derive function"""

import mongomock

from dwm import Dwm


# Setup mongomock db
DB = mongomock.MongoClient().db


# Setup Dwm instance
FIELDS = {
    "field1": {
        "lookup": [],
        "derive": [
            {
                "type": "copyValue",
                "fieldSet": ["field2"],
                "options": ["overwrite"]
            }
        ]
    }
}

DWM = Dwm(name='test', mongo=DB, fields=FIELDS)


# Let the testing begin
def test_derive_copy_value():
    """ Ensure derive copy value """
    rec = {"emailAddress": "test@test.com", "field1": "", "field2": "newvalue"}
    rec_out, _ = DWM._derive(rec, {})  # pylint: disable=W0212
    assert rec_out['field1'] == 'newvalue'



deriveIncludes = [
    {"fieldName": "field1", "deriveFieldName": "field2", "includes": "GOOD,VALUE", "excludes": "BAD,STUFF", "begins": "", "ends": "", "replace": "goodvalue"},
    {"fieldName": "field1", "deriveFieldName": "field2", "includes": "", "excludes": "", "begins": "ABC", "ends": "", "replace": "goodvalue"},
    {"fieldName": "field1", "deriveFieldName": "field2", "includes": "", "excludes": "", "begins": "", "ends": "XYZ", "replace": "goodvalue"},
    {"fieldName": "field100", "deriveFieldName": "field101"}
]



""" test_dwm_init """

import mongomock
from nose.tools import raises

from dwm import Dwm

# Setup mongomock db

DB = mongomock.MongoClient().db

# Initialization tests


def test_dwm_initialize():
    """ Test that Dwm class initializes with proper object type """
    assert isinstance(Dwm(name='test', mongo=DB), Dwm)


def test_dwm_init_name():
    """ test Dwm class sets name variable """
    assert Dwm(name='test', mongo=DB).name == 'test'


def test_dwm_init_mongo():
    """ test Dwm class sets MongoDB client """
    assert Dwm(name='test', mongo=DB).mongo == DB


def test_dwm_init_default_fields():
    """ test Dwm class initializes with default fields empty """
    assert Dwm(name='test', mongo=DB).fields == {}


def test_dwm_init_default_udf():
    """ test Dwm class initializes with default fields empty """
    assert Dwm(name='test', mongo=DB).udfs == {}

# Initialize with field settings


def test_dwm_init_fields():
    """ test Dwm class initializes with defined field set """
    fields = {
        'field1': {
            'lookup': ['genericLookup', 'genericRegex', 'fieldSpecificRegex',
                       'fieldSpecificLookup', 'normLookup', 'normIncludes'],
            'derive': [
                {
                    'type': 'deriveIncludes',
                    'fieldSet': ['field2'],
                    'options': []
                }
            ]
        }
    }

    assert Dwm(name='test', mongo=DB, fields=fields).fields == fields


@raises(ValueError)
def test_dwm_init_fields_badlookup():
    """ test Dwm class raises error with bad lookup type """
    fields = {
        'field1': {
            'lookup': ['genericLookup', 'genericRegex', 'fieldSpecificRegex',
                       'fieldSpecificLookup', 'normLookup', 'badlookup'],
            'derive': [
                {
                    'type': 'deriveIncludes',
                    'fieldSet': ['field2'],
                    'options': []
                }
            ]
        }
    }

    Dwm(name='test', mongo=DB, fields=fields)


@raises(ValueError)
def test_dwm_init_fields_badderive():
    """ test Dwm class raises error with bad derive type """
    fields = {
        'field1': {
            'lookup': ['genericLookup', 'genericRegex', 'fieldSpecificRegex',
                       'fieldSpecificLookup', 'normLookup', 'normIncludes'],
            'derive': [
                {
                    'type': 'badderive',
                    'fieldSet': ['field2'],
                    'options': []
                }
            ]
        }
    }

    Dwm(name='test', mongo=DB, fields=fields)


@raises(ValueError)
def test_dwm_init_fields_badopt():
    """ test Dwm class raises error with bad derive option type """
    fields = {
        'field1': {
            'lookup': ['genericLookup', 'genericRegex', 'fieldSpecificRegex',
                       'fieldSpecificLookup', 'normLookup', 'normIncludes'],
            'derive': [
                {
                    'type': 'deriveIncludes',
                    'fieldSet': ['field2'],
                    'options': ['badoption']
                }
            ]
        }
    }

    Dwm(name='test', mongo=DB, fields=fields)


# Initialize with User-Defined Functions

def udf_example_good(data, hist):
    """ example UDF function for testing; has expected parameters """
    return data, hist


def test_dwm_init_udf():
    """ test Dwm class initializes with defined UDFs """
    udf_set = {
        'beforeGenericValLookup': udf_example_good,
        'beforeGenericValRegex': udf_example_good,
        'beforeFieldSpecificLookup': udf_example_good,
        'beforeFieldSpecificRegex': udf_example_good,
        'beforeNormLookup': udf_example_good,
        'beforeNormRegex': udf_example_good,
        'beforeNormIncludes': udf_example_good,
        'beforeDerive': udf_example_good,
        'afterAll': udf_example_good
    }

    assert Dwm(name='test', mongo=DB, udfs=udf_set).udfs == udf_set


@raises(ValueError)
def test_dwm_init_udf_badpos():
    """ test Dwm class raises error with invalid position """

    udf_set = {
        'badposition': udf_example_good
    }

    Dwm(name='test', mongo=DB, udfs=udf_set)



"""
    dwm package setup
"""

from __future__ import print_function
from setuptools import setup, find_packages

__version__ = '1.0.2'

def readme():
    """ open readme for long_description """

    try:
        with open('README.md') as fle:
            return fle.read()
    except IOError:
        return ''

setup(
    name='dwm',
    version=__version__,
    url='https://github.com/rh-marketingops/dwm',
    license='GNU General Public License',
    author='Jeremiah Coleman',
    tests_require=['nose', 'mongomock>=3.5.0'],
    install_requires=['pymongo>=3.2.2', 'tqdm>=4.8.4'],
    author_email='colemanja91@gmail.com',
    description='Best practices for marketing data quality management',
    long_description=readme(),
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    test_suite='nose.collector',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
        ],
    keywords='marketing automation data quality cleanse washing cleaning'
)



import mongomock
from nose.tools import *
from mock import patch
from dwm import dwmAll

from .test_genericLookup import genericLookup
from .test_fieldSpecificLookup import fieldSpecificLookup
from .test_normLookup import normLookup

from .test_genericRegex import genericRegex
from .test_fieldSpecificRegex import fieldSpecificRegex
from .test_normRegex import normRegex

from .test_normIncludes import normIncludes
from .test_deriveIncludes import deriveIncludes

from .test_deriveValue import deriveValue
from .test_deriveRegex import deriveRegex

from dwm.test.test_configs import configs
import dwm.test.test_records as test_records
import dwm.test.test_udf as test_udf
from dwm.test.test_udf import ex_udf, sort_udf_1, sort_udf_2
import dwm.cleaning as cleaning
import dwm.helpers as helpers

## Initialize pre-test mongomock

db = mongomock.MongoClient().db

for row in genericLookup:

    db.genericLookup.insert_one(row)

for row in fieldSpecificLookup:

    db.fieldSpecificLookup.insert_one(row)

for row in normLookup:

    db.normLookup.insert_one(row)

for row in genericRegex:

    db.genericRegex.insert_one(row)

for row in fieldSpecificRegex:

    db.fieldSpecificRegex.insert_one(row)

for row in normRegex:

    db.normRegex.insert_one(row)

for row in normIncludes:

    db.normIncludes.insert_one(row)

for row in deriveValue:

    db.deriveValue.insert_one(row)

for row in deriveRegex:

    db.deriveRegex.insert_one(row)

for row in deriveIncludes:

    db.deriveIncludes.insert_one(row)

for row in configs:

    db.config.insert_one(row)

######################################
## Test lookups

# Test verbosity

def test_verbose():

    dataOut = dwmAll(data = test_records.record_lookupAll_genericLookup_caught, db = db, configName='test_lookupAll_genericLookup', verbose=True)
    assert dataOut[0]['field1'] == ''

# genericLookup

def test_lookupAll_genericLookup_caught():

    dataOut = dwmAll(data = test_records.record_lookupAll_genericLookup_caught, db = db, configName='test_lookupAll_genericLookup')
    assert dataOut[0]['field1'] == ''

def test_lookupAll_genericLookup_uncaught():

    dataOut = dwmAll(data = test_records.record_lookupAll_genericLookup_uncaught, db = db, configName='test_lookupAll_genericLookup')
    assert dataOut[0]['field1'] != ''

def test_lookupAll_genericLookup_notChecked():

    dataOut = dwmAll(data = test_records.record_lookupAll_genericLookup_notChecked, db = db, configName='test_lookupAll_genericLookup')
    assert dataOut[0]['field2'] != ''

# fieldSpecificLookup

def test_lookupAll_fieldSpecificLookup_caught():

    dataOut = dwmAll(data = test_records.record_lookupAll_fieldSpecificLookup_caught, db = db, configName='test_lookupAll_fieldSpecificLookup')
    assert dataOut[0]['field1'] == ''

def test_lookupAll_fieldSpecificLookup_uncaught():

    dataOut = dwmAll(data = test_records.record_lookupAll_fieldSpecificLookup_uncaught, db = db, configName='test_lookupAll_fieldSpecificLookup')
    assert dataOut[0]['field1'] != ''

def test_lookupAll_fieldSpecificLookup_notChecked():

    dataOut = dwmAll(data = test_records.record_lookupAll_fieldSpecificLookup_notChecked, db = db, configName='test_lookupAll_fieldSpecificLookup')
    assert dataOut[0]['field2'] != ''

# normLookup

def test_lookupAll_normLookup_caught():

    # do stuff for testing
    dataOut = dwmAll(data = test_records.record_lookupAll_normLookup_caught, db = db, configName='test_lookupAll_normLookup')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_lookupAll_normLookup_uncaught():

    dataOut = dwmAll(data = test_records.record_lookupAll_normLookup_uncaught, db = db, configName='test_lookupAll_normLookup')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_lookupAll_normLookup_notChecked():

    dataOut = dwmAll(data = test_records.record_lookupAll_normLookup_notChecked, db = db, configName='test_lookupAll_normLookup')
    assert dataOut[0]['field2'] != 'goodvalue'

##############################
## Test regex
# genericRegex

def test_regexAll_genericRegex_caught():

    dataOut = dwmAll(data = test_records.record_regexAll_genericRegex_caught, db = db, configName='test_regexAll_genericRegex')
    assert dataOut[0]['field1'] == ''

def test_regexAll_genericRegex_uncaught():

    dataOut = dwmAll(data = test_records.record_regexAll_genericRegex_uncaught, db = db, configName='test_regexAll_genericRegex')
    assert dataOut[0]['field1'] != ''

def test_regexAll_genericRegex_notChecked():

    dataOut = dwmAll(data = test_records.record_regexAll_genericRegex_notChecked, db = db, configName='test_regexAll_genericRegex')
    assert dataOut[0]['field2'] != ''

# fieldSpecificRegex

def test_regexAll_fieldSpecificRegex_caught():

    dataOut = dwmAll(data = test_records.record_regexAll_fieldSpecificRegex_caught, db = db, configName='test_regexAll_fieldSpecificRegex')
    assert dataOut[0]['field1'] == ''

def test_regexAll_fieldSpecificRegex_uncaught():

    dataOut = dwmAll(data = test_records.record_regexAll_fieldSpecificRegex_uncaught, db = db, configName='test_regexAll_fieldSpecificRegex')
    assert dataOut[0]['field1'] != ''

def test_regexAll_fieldSpecificRegex_notChecked():

    dataOut = dwmAll(data = test_records.record_regexAll_fieldSpecificRegex_notChecked, db = db, configName='test_regexAll_fieldSpecificRegex')
    assert dataOut[0]['field2'] != ''

# normRegex

def test_regexAll_normRegex_caught():

    dataOut = dwmAll(data = test_records.record_regexAll_normRegex_caught, db = db, configName='test_regexAll_normRegex')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_regexAll_normRegex_uncaught():

    # do stuff for testing
    dataOut = dwmAll(data = test_records.record_regexAll_normRegex_uncaught, db = db, configName='test_regexAll_normRegex')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_regexAll_normRegex_notChecked():

    dataOut = dwmAll(data = test_records.record_regexAll_normRegex_notChecked, db = db, configName='test_regexAll_normRegex')
    assert dataOut[0]['field2'] != 'goodvalue'

# normIncludes

def test_normIncludes_included_caught():

    dataOut = dwmAll(data = test_records.record_normIncludes_included_caught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_normIncludes_included_uncaught():

    dataOut = dwmAll(data = test_records.record_normIncludes_included_uncaught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_normIncludes_excluded_caught():

    dataOut = dwmAll(data = test_records.record_normIncludes_excluded_caught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_normIncludes_excluded_uncaught():

    dataOut = dwmAll(data = test_records.record_normIncludes_excluded_uncaught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_normIncludes_begins_caught():

    dataOut = dwmAll(data = test_records.record_normIncludes_begins_caught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_normIncludes_begins_uncaught():

    dataOut = dwmAll(data = test_records.record_normIncludes_begins_uncaught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_normIncludes_ends_caught():

    dataOut = dwmAll(data = test_records.record_normIncludes_ends_caught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_normIncludes_ends_uncaught():

    dataOut = dwmAll(data = test_records.record_normIncludes_ends_uncaught, db = db, configName='test_normIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_normIncludes_notChecked():

    dataOut = dwmAll(data = test_records.record_normIncludes_notChecked, db = db, configName='test_normIncludes')
    assert dataOut[0]['field2'] != 'goodvalue'


###################################
## Test derive
# deriveValue

def test_deriveAll_deriveValue_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveValue_caught, db = db, configName='test_deriveAll_deriveValue')
    assert dataOut[0]['field1'] == 'newvalue'

def test_deriveAll_deriveValue_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveValue_uncaught, db = db, configName='test_deriveAll_deriveValue')
    assert dataOut[0]['field1'] != 'newvalue'

def test_deriveAll_deriveValue_notChecked():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveValue_notChecked, db = db, configName='test_deriveAll_deriveValue')
    assert dataOut[0]['field3'] == ''

def test_deriveAll_deriveValue_overwriteFalse():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveValue_overwriteFalse, db = db, configName='test_deriveAll_deriveValue_overwriteFalse')
    assert dataOut[0]['field1'] == 'oldvalue'

def test_deriveAll_deriveValue_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveValue_blankIfNoMatch, db = db, configName='test_deriveAll_deriveValue_blankIfNoMatch')
    assert dataOut[0]['field1'] == ''

# copyValue

def test_deriveAll_copyValue():

    dataOut = dwmAll(data = test_records.record_deriveAll_copyValue, db = db, configName='test_deriveAll_copyValue')
    assert dataOut[0]['field1'] == 'newvalue'

# deriveRegex

def test_deriveAll_deriveRegex_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveRegex_caught, db = db, configName='test_deriveAll_deriveRegex')
    assert dataOut[0]['field1'] == 'newvalue'

def test_deriveAll_deriveRegex_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveRegex_uncaught, db = db, configName='test_deriveAll_deriveRegex')
    assert dataOut[0]['field1'] != 'newvalue'

def test_deriveAll_deriveRegex_notChecked():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveRegex_notChecked, db = db, configName='test_deriveAll_deriveRegex')
    assert dataOut[0]['field3'] == ''

def test_deriveAll_deriveRegex_overwriteFalse():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveRegex_overwriteFalse, db = db, configName='test_deriveAll_deriveRegex_overwriteFalse')
    assert dataOut[0]['field1'] == 'oldvalue'

def test_deriveAll_deriveRegex_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveRegex_blankIfNoMatch, db = db, configName='test_deriveAll_deriveRegex_blankIfNoMatch')
    assert dataOut[0]['field1'] == ''

# deriveIncludes

def test_deriveAll_deriveIncludes_included_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_included_caught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_deriveAll_deriveIncludes_included_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_included_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_deriveAll_deriveIncludes_excluded_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_excluded_caught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_deriveAll_deriveIncludes_excluded_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_excluded_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_deriveAll_deriveIncludes_begins_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_begins_caught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_deriveAll_deriveIncludes_begins_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_begins_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_deriveAll_deriveIncludes_ends_caught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_ends_caught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] == 'goodvalue'

def test_deriveAll_deriveIncludes_ends_uncaught():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_ends_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field1'] != 'goodvalue'

def test_deriveAll_deriveIncludes_notChecked():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_notChecked, db = db, configName='test_deriveAll_deriveIncludes')
    assert dataOut[0]['field2'] != 'goodvalue'

def test_deriveAll_deriveIncludes_overwriteFalse():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_overwriteFalse, db = db, configName='test_deriveAll_deriveIncludes_overwriteFalse')
    assert dataOut[0]['field1'] == 'oldvalue'

def test_deriveAll_deriveIncludes_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.record_deriveAll_deriveIncludes_blankIfNoMatch, db = db, configName='test_deriveAll_deriveIncludes_blankIfNoMatch')
    assert dataOut[0]['field1'] == ''

# ensure proper sorting on derive
def test_derive_sort():

    dataOut = dwmAll(data = test_records.record_derive_sort, db = db, configName='test_derive_sort')
    assert dataOut[0]['field1'] == 'correctvalue'

# ensure derive check match
def test_deriveIncludes_deriveCheckMatch():
    dataOut_FirstExecution = dwmAll(data=test_records.history_deriveIncludes_deriveCheckMatch, db=db, configName='test_deriveAll_deriveIncludes_deriveCheckMatch')
    dataOut_SecondExecution = dwmAll(data = dataOut_FirstExecution, db = db, configName='test_deriveAll_deriveIncludes_deriveCheckMatch')
    assert dataOut_FirstExecution[0]['field1'] == dataOut_SecondExecution[0]['field1']

######################################
## test contact history

# genericLookup

def test_history_genericLookup_caught():

    dataOut = dwmAll(data = test_records.history_genericLookup_caught, db = db, configName='test_lookupAll_genericLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['genericLookup']['from'] == 'badvalue'

def test_history_genericLookup_uncaught():

    dataOut = dwmAll(data = test_records.history_genericLookup_uncaught, db = db, configName='test_lookupAll_genericLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_genericLookup_notChecked():

    dataOut = dwmAll(data = test_records.history_genericLookup_notChecked, db = db, configName='test_lookupAll_genericLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# # fieldSpecificLookup

def test_history_fieldSpecificLookup_caught():

    dataOut = dwmAll(data = test_records.history_fieldSpecificLookup_caught, db = db, configName='test_lookupAll_fieldSpecificLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['fieldSpecificLookup']['from'] == 'badvalue'

def test_history_fieldSpecificLookup_uncaught():

    dataOut = dwmAll(data = test_records.history_fieldSpecificLookup_uncaught, db = db, configName='test_lookupAll_fieldSpecificLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_fieldSpecificLookup_notChecked():

    dataOut = dwmAll(data = test_records.history_fieldSpecificLookup_notChecked, db = db, configName='test_lookupAll_fieldSpecificLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()


# # CurrentFieldLookup
def test_writeContactHistory_currentField():

    dataOut = dwmAll(data = test_records.record_writeContactHistory_writeConfig, db = db, configName='test_writeContactHistory_writeConfig')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['_current'] == 0

def test_writeContactHistory_currentFieldInc():
    # Drop contactHistory before new UT
    db.contactHistory.drop()
    # Create ContactHistory with _currentField
    dataOutOldRec = dwmAll(data=test_records.record_writeContactHistory_historyCurrent1, db=db, configName='test_writeContactHistory_writeConfig')

    # Create new ContactHistory with _currentField and increment old record.
    dataOutNextRec = dwmAll(data=test_records.record_writeContactHistory_historyCurrent2, db=db, configName='test_writeContactHistory_writeConfig')

    # Get old rec with incremented _current
    hist = db.contactHistory.find_one({"_id": dataOutOldRec[0]['historyId']})
    assert hist['_current'] == 1

# normLookup
def test_history_normLookup_caught():

    # do stuff for testing
    dataOut = dwmAll(data = test_records.history_normLookup_caught, db = db, configName='test_lookupAll_normLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['normLookup']['from'] == 'badvalue'

def test_history_normLookup_uncaught():

    dataOut = dwmAll(data = test_records.history_normLookup_uncaught, db = db, configName='test_lookupAll_normLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_normLookup_notChecked():

    dataOut = dwmAll(data = test_records.history_normLookup_notChecked, db = db, configName='test_lookupAll_normLookup')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# normIncludes

def test_history_normIncludes_caught():

    # do stuff for testing
    dataOut = dwmAll(data = test_records.history_normIncludes_included_caught, db = db, configName='test_normIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['normIncludes']['from'] == 'findgoodinvaluejunk'

def test_history_normIncludes_uncaught():

    dataOut = dwmAll(data = test_records.history_normIncludes_included_uncaught, db = db, configName='test_normIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_normIncludes_notChecked():

    dataOut = dwmAll(data = test_records.history_normIncludes_notChecked, db = db, configName='test_normIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# genericRegex

def test_history_genericRegex_caught():

    dataOut = dwmAll(data = test_records.history_genericRegex_caught, db = db, configName='test_regexAll_genericRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['genericRegex']['from'] == 'badvalue'

def test_history_genericRegex_uncaught():

    dataOut = dwmAll(data = test_records.history_genericRegex_uncaught, db = db, configName='test_regexAll_genericRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_genericRegex_notChecked():

    dataOut = dwmAll(data = test_records.history_genericRegex_notChecked, db = db, configName='test_regexAll_genericRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# fieldSpecificRegex

def test_history_fieldSpecificRegex_caught():

    dataOut = dwmAll(data = test_records.history_fieldSpecificRegex_caught, db = db, configName='test_regexAll_fieldSpecificRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['fieldSpecificRegex']['from'] == 'badvalue'

def test_history_fieldSpecificRegex_uncaught():

    dataOut = dwmAll(data = test_records.record_regexAll_fieldSpecificRegex_uncaught, db = db, configName='test_regexAll_fieldSpecificRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_fieldSpecificRegex_notChecked():

    dataOut = dwmAll(data = test_records.history_fieldSpecificRegex_notChecked, db = db, configName='test_regexAll_fieldSpecificRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# normRegex

def test_history_normRegex_caught():

    dataOut = dwmAll(data = test_records.history_normRegex_caught, db = db, configName='test_regexAll_normRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['normRegex']['from'] == 'badvalue'

def test_history_normRegex_uncaught():

    # do stuff for testing
    dataOut = dwmAll(data = test_records.history_normRegex_uncaught, db = db, configName='test_regexAll_normRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_normRegex_notChecked():

    dataOut = dwmAll(data = test_records.history_normRegex_notChecked, db = db, configName='test_regexAll_normRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field2' not in hist.keys()

# deriveValue

def test_history_deriveValue_caught():

    dataOut = dwmAll(data = test_records.history_deriveValue_caught, db = db, configName='test_deriveAll_deriveValue')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveValue']['from'] == ''

def test_history_deriveValue_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveValue_uncaught, db = db, configName='test_deriveAll_deriveValue')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveValue_notChecked():

    dataOut = dwmAll(data = test_records.history_deriveValue_notChecked, db = db, configName='test_deriveAll_deriveValue')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field3' not in hist.keys()

def test_history_deriveValue_overwriteFalse():

    dataOut = dwmAll(data = test_records.history_deriveValue_overwriteFalse, db = db, configName='test_deriveAll_deriveValue_overwriteFalse')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveValue_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.history_deriveValue_blankIfNoMatch, db = db, configName='test_deriveAll_deriveValue_blankIfNoMatch')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveValue']['using']['blankIfNoMatch'] == 'no match found'

# copyValue

def test_history_copyValue():

    dataOut = dwmAll(data = test_records.history_copyValue, db = db, configName='test_deriveAll_copyValue')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['copyValue']['from'] == ''

# deriveRegex

def test_history_deriveRegex_caught():

    dataOut = dwmAll(data = test_records.history_deriveRegex_caught, db = db, configName='test_deriveAll_deriveRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveRegex']['from'] == ''

def test_history_deriveRegex_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveRegex_uncaught, db = db, configName='test_deriveAll_deriveRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveRegex_notChecked():

    dataOut = dwmAll(data = test_records.history_deriveRegex_notChecked, db = db, configName='test_deriveAll_deriveRegex')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field3' not in hist.keys()

def test_history_deriveRegex_overwriteFalse():

    dataOut = dwmAll(data = test_records.history_deriveRegex_overwriteFalse, db = db, configName='test_deriveAll_deriveRegex_overwriteFalse')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveRegex_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.history_deriveRegex_blankIfNoMatch, db = db, configName='test_deriveAll_deriveRegex_blankIfNoMatch')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveRegex']['pattern'] == 'no matching pattern'

# deriveIncludes

def test_history_deriveIncludes_included_caught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_included_caught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveIncludes']['from'] == ''

def test_history_deriveIncludes_included_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_included_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveIncludes_excluded_caught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_excluded_caught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveIncludes_excluded_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_excluded_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveIncludes']['from'] == ''

def test_history_deriveIncludes_begins_caught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_begins_caught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveIncludes']['from'] == ''

def test_history_deriveIncludes_begins_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_begins_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveIncludes_ends_caught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_ends_caught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveIncludes']['from'] == ''

def test_history_deriveIncludes_ends_uncaught():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_ends_uncaught, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveIncludes_notChecked():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_notChecked, db = db, configName='test_deriveAll_deriveIncludes')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field3' not in hist.keys()

def test_history_deriveIncludes_overwriteFalse():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_overwriteFalse, db = db, configName='test_deriveAll_deriveIncludes_overwriteFalse')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert 'field1' not in hist.keys()

def test_history_deriveIncludes_blankIfNoMatch():

    dataOut = dwmAll(data = test_records.history_deriveIncludes_blankIfNoMatch, db = db, configName='test_deriveAll_deriveIncludes_blankIfNoMatch')
    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})
    assert hist['field1']['deriveIncludes']['using']['blankIfNoMatch'] == 'no match found'

#########################################
## Test options around history
# returnHistoryId False

def test_returnHistoryId_False():

    dataOut = dwmAll(data = test_records.record_returnHistoryId_False, db = db, configName='test_returnHistoryId_False')
    assert 'historyId' not in dataOut[0].keys()

def test_writeContactHistory_False():

    dataOut = dwmAll(data = test_records.record_writeContactHistory_False, db = db, configName='test_writeContactHistory_False')
    assert 'historyId' not in dataOut[0].keys()

def test_writeContactHistory_writeConfig():

    dataOut = dwmAll(data = test_records.record_writeContactHistory_writeConfig, db = db, configName='test_writeContactHistory_writeConfig')

    hist = db.contactHistory.find_one({"_id": dataOut[0]['historyId']})

    assert hist['configName'] == 'test_writeContactHistory_writeConfig'

#########################################
## configuration does not exist
@raises(Exception)
def test_configDoesNotExist():

    dataOut = dwmAll(data = test_records.record_configDoesNotExist, db = db, configName='test_configDoesNotExist')

#########################################
# userDefinedFunctions

def test_udf_beforeGenericValidation():

    dataOut = dwmAll(data = test_records.record_udf_beforeGenericValidation, db = db, configName='test_udf_beforeGenericValidation', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeGenericRegex():

    dataOut = dwmAll(data = test_records.record_udf_beforeGenericRegex, db = db, configName='test_udf_beforeGenericRegex', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeFieldSpecificValidation():

    dataOut = dwmAll(data = test_records.record_udf_beforeFieldSpecificValidation, db = db, configName='test_udf_beforeFieldSpecificValidation', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeFieldSpecificRegex():

    dataOut = dwmAll(data = test_records.record_udf_beforeFieldSpecificRegex, db = db, configName='test_udf_beforeFieldSpecificRegex', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeNormalization():

    dataOut = dwmAll(data = test_records.record_udf_beforeNormalization, db = db, configName='test_udf_beforeNormalization', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeNormalizationRegex():

    dataOut = dwmAll(data = test_records.record_udf_beforeNormalizationRegex, db = db, configName='test_udf_beforeNormalizationRegex', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeNormalizationIncludes():

    dataOut = dwmAll(data = test_records.record_udf_beforeNormalizationIncludes, db = db, configName='test_udf_beforeNormalizationIncludes', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_beforeDeriveData():

    dataOut = dwmAll(data = test_records.record_udf_beforeDeriveData, db = db, configName='test_udf_beforeDeriveData', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_afterProcessing():

    dataOut = dwmAll(data = test_records.record_udf_afterProcessing, db = db, configName='test_udf_afterProcessing', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

def test_udf_sort():

    dataOut = dwmAll(data = test_records.record_udf_sort, db = db, configName='test_udf_sort', udfNamespace=__name__)
    assert dataOut[0]['field1'] == 'goodvalue'

######################################
## Test value exceptions in base-level functions
@raises(Exception)
def test_udf_afterProcessing_invalidFcn():

    dataOut = dwmAll(data = test_records.record_udf_afterProcessing_invalidFcn, db = db, configName='test_udf_afterProcessing_invalidFcn', udfNamespace=__name__)

## bad lookup type:
@raises(ValueError)
def test_DataLookup_badType():

    testVal, histObj = cleaning.DataLookup(fieldVal='', db=db, lookupType='badlookup', fieldName='')

# bad regex type:
@raises(ValueError)
def test_RegexLookup_badType():

    testVal, histObj = cleaning.RegexLookup(fieldVal='', db=db, lookupType='badregex', fieldName='')

# multiple deriveCopy inputs:
@raises(Exception)
def test_DeriveDataCopyValue_badType():

    testVal, histObj = cleaning.DeriveDataCopyValue(fieldName='test', deriveInput={"field1": "a", "field2": "b"}, overwrite=True, fieldVal='')

# multiple deriveRegex inputs:
@raises(Exception)
def test_DeriveDataRegex_badType():

    testVal, histObj = cleaning.DeriveDataRegex(fieldName='test', db=db, deriveInput={"field1": "a", "field2": "b"}, overwrite=True, fieldVal='')

# requires config input
@raises(Exception)
def test_dwmAll_requiresConfigInput():

    dataOut = dwmAll(data = test_records.record_dwmAll_noConfig, db = db, udfNamespace=__name__)

# doesn't allow multiple config
@raises(Exception)
def test_dwmAll_requiresOneConfig():

    dataOut = dwmAll(data = test_records.record_dwmAll_noConfig, db = db, config={"configName": "testConfig"}, configName="testConfig", udfNamespace=__name__)

# doesn't allow invalid includes lookupType
@raises(ValueError)
def test_includesLookup_badType():

    testVal, histObj = cleaning.IncludesLookup(fieldVal='', lookupType='badlookup', db = db, fieldName='')

# deriveIncludes requires deriveInput
@raises(ValueError)
def test_includesLookup_requiresDeriveFieldName():

    testVal, histObj = cleaning.IncludesLookup(fieldVal='', lookupType='deriveIncludes', deriveFieldName='test', db = db, fieldName='')

# deriveIncludes requires deriveFieldName
@raises(ValueError)
def test_includesLookup_requiresDeriveFieldName():

    testVal, histObj = cleaning.IncludesLookup(fieldVal='', lookupType='deriveIncludes', deriveInput='test', db = db, fieldName='')

# warns if returned document doesn't match schema
@patch('dwm.cleaning.warnings.warn')
def test_includesLookup_normIncludes_warnOnBadSchema(mock_warnings):

    cleaning.IncludesLookup(fieldVal='1234', lookupType='normIncludes', db = db, fieldName='field100')
    assert(mock_warnings.called)

# warns if returned document doesn't match schema
@patch('dwm.cleaning.warnings.warn')
def test_includesLookup_deriveIncludes_warnOnBadSchema(mock_warnings):

    cleaning.IncludesLookup(fieldVal='', lookupType='deriveIncludes', db = db, fieldName='field100', deriveFieldName='field101', deriveInput={"field101": "abcd"})
    assert(mock_warnings.called)

# warns if returned document doesn't match schema
@patch('dwm.cleaning.warnings.warn')
def test_regexLookup_warnOnBadSchema(mock_warnings):

    cleaning.RegexLookup(fieldVal='', db = db, fieldName="field100", lookupType="fieldSpecificRegex")
    assert(mock_warnings.called)

# warns if returned document doesn't match schema
@patch('dwm.cleaning.warnings.warn')
def test_deriveValue_warnOnBadSchema(mock_warnings):

    cleaning.DeriveDataLookup(fieldName="field100", db = db, deriveInput = {"field101": "abc"})
    assert(mock_warnings.called)

# warns if returned document doesn't match schema
@patch('dwm.cleaning.warnings.warn')
def test_deriveRegex_warnOnBadSchema(mock_warnings):

    cleaning.DeriveDataRegex(fieldName="field100", db = db, deriveInput = {"field101": "abc"}, overwrite=False, fieldVal='')
    assert(mock_warnings.called)



def ex_udf(data, histObj):

    data['field1'] = 'goodvalue'

    return data, histObj

def sort_udf_1(data, histObj):

    data['field1'] = 'badvalue'

    return data, histObj

def sort_udf_2(data, histObj):

    data['field1'] = 'anotherbadvalue'

    return data, histObj



""" test apply udfs ( user copy value case of derive function"""

import mongomock

from nose.tools import raises

from dwm import Dwm

from dwm.test.test_udf import ex_udf
from dwm.test.test_udf import sort_udf_1
from dwm.test.test_udf import sort_udf_2

# Setup mongomock db
DB = mongomock.MongoClient().db


def setup_dwm_instance(name_str, mongo_obj, udfs_dict):
    """
    set up DWM instance
    :param str name_str: Name of configuration
    :param MongoClient mongo: MongoDB connection
    :param dict udfs: dict of udfs to run
    """

    return Dwm(name=name_str, mongo=mongo_obj, udfs=udfs_dict)


# Let the testing begin
def test_udf_before_generic_validation():
    """ Ensure udf before generic validation """
    # Setup Dwm instance

    udfs_obj = {
        "beforeGenericValLookup": {
            "1": ex_udf
        },
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeGenericValLookup') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_generic_regex():
    """ Ensure udf before generic regex """
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {
            "1": ex_udf
        },
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeGenericValRegex') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_field_specific_validation():
    """ Ensure udf before field specific validation """
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {
            "1": ex_udf
        },
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeFieldSpecificLookup') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_field_specific_regex():
    """ Ensure udf before field specific regex """
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {
            "1": ex_udf
        },
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeFieldSpecificRegex') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_normalization():
    """ Ensure udf before normalization """
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {
            "1": ex_udf
        },
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeNormLookup') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_normalization_regex():
    """ Ensure udf before normalization regex"""
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {
            "1": ex_udf
        },
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeNormRegex') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_normalization_includes():
    """ Ensure udf before normalization includes"""
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {
            "1": ex_udf
        },
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeNormIncludes') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_before_derive_data():
    """ Ensure udf before derive data"""
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {
            "1": ex_udf
        },
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeDerive') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_after_all():
    """ Ensure udf before derive data"""
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {
            "1": ex_udf
        }
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'afterAll') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


def test_udf_sort():
    """ Ensure udf sort case"""
    udfs_obj = {
        "beforeGenericValLookup": {
            "3": ex_udf,
            "1": sort_udf_1,
            "2": sort_udf_2
        },
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {}
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    rec_out, _ = dwm_obj._apply_udfs(rec, {}, 'beforeGenericValLookup') # pylint: disable=W0212
    assert rec_out['field1'] == 'goodvalue'


@raises(Exception)
def test_udf_after_all_invalid_func():
    """ Test value exceptions in base-level functions """
    udfs_obj = {
        "beforeGenericValLookup": {},
        "beforeGenericValRegex": {},
        "beforeFieldSpecificLookup": {},
        "beforeFieldSpecificRegex": {},
        "beforeNormLookup": {},
        "beforeNormRegex": {},
        "beforeNormIncludes": {},
        "beforeDerive": {},
        "afterAll": {
            "1": "bad_udf"
        }
    }

    dwm_obj = setup_dwm_instance('test', DB, udfs_obj)

    rec = {"emailAddress": "test@test.com", "field1": ""}
    _, _ = dwm_obj._apply_udfs(rec, {}, 'afterAll') # pylint: disable=W0212
